#!/usr/bin/env python3

import uproot
from .histogram import histogram1d, histogram2d

def readroot(rootfile, histpath, variation_paths={}, ignore_missing_variations=False, **kwargs):
    """Reads a histogram (possibly with systematic variations) from a ROOT file.

    :param rootfile: path to ROOT file
    :type rootfile: ``str``
    :param histopath: path of the histogram inside the file
    :type histopath: ``str``
    :param variation_paths: optional dictionary of variation names (keys) and
        paths to the variation histograms inside the same ROOT file (values).
        NOTE: we can easily add the ability to read variation histograms also
        from other ROOT files than the nominal, get in touch if you want that.
    :type variation_paths: ``dict`` of ``str : str``
    :param **kwargs: get passed on to histogram constructor. An important one is
        ``areas=False`` if retrieving ratios (e.g. efficiencies).
    """

    # TODO add this:
    #
    # The variation paths may contain the following wildcards:
    # - "@" will be interpreted as a numerical index starting at 0 and continuing until no more histograms are found
    # - "#" is the same as "@", but starting at 1
    # - Arbitrary Unix-style wildcards (see Python module fnmatch). These may not be combined with the above
    # custom wildcards. Note: including Unix-style wildcards may be very slow.

    file = uproot.open(rootfile)
    urhist = file[histpath]

    if not any(['TH1' in urhist.classname, 'TH2' in urhist.classname]):
        raise TypeError('Can only read ROOT.TH1 and ROOT.TH2 objects, but '
            f'found {urhist.classname} object in {rootfile} at {histpath}')

    binedges = tuple([axis.edges() for axis in urhist.axes])
    if len(binedges) == 1:
        binedges = binedges[0]
    corr_variations = {}
    for name, path in variation_paths.items():
        try:
            corr_variations[name] = file[path].values()
        except ReferenceError:
            if not ignore_missing_variations:
                raise ValueError('While reading variation histograms, could '
                    f'not read object object in {rootfile} at {path}')
    attributes = {
        'name' : urhist.name,
        'title' : urhist.title,
        'provenance' : f'{rootfile}:{histpath}'
    }
    histclass = histogram1d if 'TH1' in urhist.classname else histogram2d
    kwargs['areas'] = kwargs.get('areas', True)
    return histclass(binedges, urhist.values(),
        uncorr_variations={'Statistical__1up' : urhist.values() + urhist.errors(),
        'Statistical__1down' : urhist.values() - urhist.errors()},
        corr_variations=corr_variations, attributes=attributes,
        plot_attributes={'label' : kwargs.get('name', urhist.name)}, **kwargs)



# Auxiliary modules
from os.path import join, dirname
import sys

# Module to be tested
sys.path.append(join(dirname(__file__), '..'))
import heppy
from heppy.heatmap import _make_text



#
# Set up
#
_values_for_some_bin = {
	'nominal' : 11.1,
	'uncert_up' : 0.5,
	'uncert_down' : -0.7,
	'stat_up' : 0.2,
	'stat_down' : -0.1,
	'syst_up' : 0.3,
	'syst_down' : -0.4,
}





def test_to_str():
	'''
	Test formatting a single float value to a string with a given precision
	'''
	from heppy import TextFormatter
	test_value = 1234.345
	assert TextFormatter._to_str(test_value) == '1230'
	assert TextFormatter._to_str(test_value, significants=1) == '1000'
	assert TextFormatter._to_str(test_value, significants=2) == '1200'
	assert TextFormatter._to_str(test_value, significants=3) == '1230'
	assert TextFormatter._to_str(test_value, significants=4) == '1234'
	assert TextFormatter._to_str(test_value, significants=5) == '1234.3'
	assert TextFormatter._to_str(test_value, significants=6) == '1234.35'
	assert TextFormatter._to_str(test_value, significants=7) == '1234.345'
	assert TextFormatter._to_str(test_value, significants=8) == '1234.345'
	assert TextFormatter._to_str(test_value, fixedprec=4) == '1234.345'
	assert TextFormatter._to_str(test_value, fixedprec=3) == '1234.345'
	assert TextFormatter._to_str(test_value, fixedprec=2) == '1234.35'
	assert TextFormatter._to_str(test_value, fixedprec=1) == '1234.3'
	assert TextFormatter._to_str(test_value, fixedprec=0) == '1234'
	assert TextFormatter._to_str(test_value, fixedprec=-1) == '1230'
	assert TextFormatter._to_str(test_value, fixedprec=-2) == '1200'
	assert TextFormatter._to_str(test_value, fixedprec=-3) == '1000'
	assert TextFormatter._to_str(test_value, fixedprec=-4) == '0'



def test_make_text_with_string_template():
	'''
	Test formatting of heatmap bin content text with a string template
	'''
	template = '{nominal} {uncert_up} {uncert_down} {stat_up} {stat_down} {syst_up} {syst_down}'
	text = _make_text(template, _values_for_some_bin)
	assert text == '11.1 0.5 -0.7 0.2 -0.1 0.3 -0.4'



def test_make_text_with_named_formatting_function():
	'''
	Test formatting of heatmap bin content text with a named formatting function
	'''
	def template_function(nominal=None, uncert_up=None, uncert_down=None, stat_up=None, stat_down=None, syst_up=None, syst_down=None):
		return '_'.join([str(arg) for arg in [nominal, uncert_up, uncert_down, stat_up, stat_down, syst_up, syst_down]])
	text = _make_text(template_function, _values_for_some_bin)
	assert text == '11.1_0.5_-0.7_0.2_-0.1_0.3_-0.4'



def test_make_text_with_lambda_formatting_function():
	'''
	Test formatting of heatmap bin content text with an unnamed formatting function (lambda)
	'''
	template_lambda = lambda nominal, uncert_up, uncert_down, stat_up, stat_down, syst_up, syst_down : ':'.join([str(arg) for arg in [nominal, uncert_up, uncert_down, stat_up, stat_down, syst_up, syst_down]])
	text = _make_text(template_lambda, _values_for_some_bin)
	assert text == '11.1:0.5:-0.7:0.2:-0.1:0.3:-0.4'



from copy import deepcopy
import numpy as np

class Value(object):
    """A single value with uncertainties.

    :param nominal: nominal value
    :type nominal: ``float``
    :param uncorr_variations: dictionary of variations that are uncorrelated between different :py:class:`heppy.value` objects even when they have the same key
    :type uncorr_variations: ``dict`` of ``str`` and ``float``
    :param corr_variations: dictionary of variations that are fully correlated between different :py:class:`heppy.value` objects when they have the same key, and uncorrelated otherwise
    :type corr_variations: ``dict`` of ``str`` and ``float``
    :param attributes: dictionary of completely arbitrary attributes that the user can provide/change/access. E.g. information about the data sample that produced the value
    :type attributes: :code:`dict`
    """
    def __init__(self, nominal, uncorr_variations={}, corr_variations={}, attributes={}):
        super(Value, self).__init__()
        self.nominal = nominal
        self.uncorr_variations = deepcopy(uncorr_variations)
        self.corr_variations = deepcopy(corr_variations)
        self.attributes = deepcopy(attributes)



    def to_atlasiff(self, attributes={}, up_suffix='__1up', down_suffix='__1down'):
        """Returns string representation in ATLAS IFF format.

        This is the XML format used by the fake-lepton background tool of the
        ATLAS Isolation and Fakes Forum.

        :param attributes: dictionary of attributes to put in the bin-tag
        :type attributes: ``dict``
        :param up_suffix: suffix in variation keys to designate an up variation
        :type attributes: ``str``
        :param down_suffix: suffix in variation keys to designate an down variation
        :type attributes: ``str``

        Usage example:

        .. code-block:: python

            >>> import heppy as hp
            >>> nominal = 12.3
            >>> uncorr_variations = {
                    'stat__1up' : 12.4,
                    'stat__1down' : 12.1,
                    }
            >>> corr_variations={
                    'efficiency__1up' : 13.1,
                    'efficiency__1down' : 9.8,
                    'energy_scale__1up' : 10.5,
                    }
            >>> v = hp.Value(nominal, uncorr_variations=uncorr_variations, corr_variations=corr_variations)
            >>> v.to_atlasiff(attributes={'pt' : '[20,inf]', '|eta|' : '[0.0,0.6]'})
            '<bin pt="[20,inf]" |eta|="[0.0,0.6]"> 12.3 +0.1-0.2 (stat) -1.8+0.0 (energy_scale) +0.8-2.5 (efficiency) </bin>'

        """
        attribute_string = ' '.join([f'{key}="{val}"' for key, val in attributes.items()])
        stripped_uncorr_keys = set([key.replace(up_suffix, '').replace(down_suffix, '') for key in self.uncorr_variations.keys()])
        uncorr_string = ' '.join([f'{np.format_float_positional(self.uncorr_variations.get(stripped_key+up_suffix, self.nominal)-self.nominal, precision=7, sign=True)}{np.format_float_positional(self.uncorr_variations.get(stripped_key+down_suffix, self.nominal)-self.nominal, precision=7, sign=True)} ({stripped_key})' for stripped_key in stripped_uncorr_keys])
        stripped_corr_keys = set([key.replace(up_suffix, '').replace(down_suffix, '') for key in self.corr_variations.keys()])
        corr_string = ' '.join([f'{np.format_float_positional(self.corr_variations.get(stripped_key+up_suffix, self.nominal)-self.nominal, precision=7, sign=True)}{np.format_float_positional(self.corr_variations.get(stripped_key+down_suffix, self.nominal)-self.nominal, precision=7, sign=True)} ({stripped_key})' for stripped_key in stripped_corr_keys])
        return f'<bin {attribute_string}> {np.format_float_positional(self.nominal, precision=7)} {uncorr_string} {corr_string} </bin>'



    def net_variations(self, variations='all', subtract_nominal=False, relative=False):
        '''
        Return upper and lower net height variation of the value.

        @variations should be a sequence of considered variation names or the string 'all'
        @subtract_nominal: if True, return the differences with respect to the nominal heights
        @relative: if True, divide by the nominal heights

        CAUTION: this method cannot yet deal with systematic uncertainties for which the up- and down-shift lie on the same side of the nominal.
        This is because the variations are fundamentally treated independently of each other, so there is no sense of the up- and down-shift
        being related to the same underlying uncertainty source.
        '''
        all_variations = {**self.uncorr_variations, **self.corr_variations} # correlated/uncorrelated only matters for rebinning so far...
        if variations == 'all':
            variations = all_variations.keys()
        if not variations:
            raise RuntimeError('Please specify some variations to compute net variations!')
        v = np.array([all_variations[name] for name in all_variations if name in variations])
        #print('variations array')
        #print(v)
        shifts = v - self.nominal
        lower = self.nominal - np.sqrt(np.sum(shifts.clip(max=0)**2, axis=0)) # add negative shifts from nominal in quadrature
        #print('lower')
        #print(lower)
        upper = self.nominal + np.sqrt(np.sum(shifts.clip(min=0)**2, axis=0)) # add positive shifts from nominal in quadrature
        #print('upper')
        #print(upper)
        if subtract_nominal:
            lower = lower - self.nominal
            upper = upper - self.nominal
        if relative:
            # If height == 0, division would give NaN. Using nan_to_num, this is replaced by zero.
            lower = np.nan_to_num(lower / self.nominal)
            upper = np.nan_to_num(upper / self.nominal)
        return upper, lower












class panel(object):
    '''
    A panel holds histograms and other information that describe a panel of a plot

    :param title: title of the panel
    :type title: `str`
    :param height: relative height of the panel with respect to any other panels in a plot
    :type height: `float`
    :param xlabel: x-axis label
    :type xlabel: `str`
    :param ylabel: y-axis label
    :type ylabel: `str`
    :param logx: plot x-axis on a logarithmic scale?
    :type logx: `bool`
    :param logy: plot y-axis on a logarithmic scale?
    :type logy: `bool`
    :param stack: histogram stack to plot
    :type stack: :py:class:`heppy.histostack`, or `None`
    :param curves: histograms to plot as curves
    :type curves: list of :py:class:`heppy.histogram1d`
    :param bands: histograms to plot as (uncertainty) bands
    :type bands: list of :py:class:`heppy.histogram1d`
    :param points: histograms to plot as points located at the centre of each of their bins
    :type points: list of :py:class:`heppy.histogram1d`
    :param pointshift: distance to shift points by horizontally to avoid overlap and improve readability. This functionality is poorly tested and may be broken
    :type pointshift: `float`
    :param scatters: this is a bit of an oddball. You can use it to plot 2D scatters that aren't really histograms. The x-values are lower bin edges (the uppermost binedge is not used for anything) while the y-values are the areas.
    :type scatters: list of :py:class:`heppy.histogram1d`
    :param ylims: can be used to manually set lower and upper y-axis limits, e.g. ylims=(0.0, 2.0)
    :type ylims: `tuple` of `float`, or `None`
    :param legend_title: legend title
    :type legend_title: `str`
    :param legend_loc: legend title
    :type legend_loc: whatever Matplotlib accepts for the ``loc`` keyword arg of legend
    :param unbinned: unbinned curves, given as tuple of x and y values (for matplotlib.pyplot.plot)
    :type unbinned: list of tuples, each tuple has two np.arrays and a plot attribute dict
    '''
    def __init__(self, title='', height=1., xlabel='', ylabel='', logx=False, logy=False, stack=None, curves=[], bands=[], points=[], pointshift=0., scatters=[], ylims=None, nolegend=False, legend_title=None, legend_loc=None, unbinned=[]):
        super(panel, self).__init__()
        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.logx = logx
        self.logy = logy
        self.stack = stack
        self.curves = curves
        self.bands = bands
        self.points = points
        self.pointshift = pointshift
        self.scatters = scatters
        self.height = height
        self.ylims = ylims
        self.nolegend = nolegend
        self.legend_title = legend_title
        self.legend_loc = legend_loc
        self.unbinned = unbinned



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='heppyness',
    license='HEP license, based on MIT license',
    version='2.0.5',
    author='Stefan Richter',
    author_email='stefan.richter@cern.ch',
    maintainer='Stefan Richter',
    maintainer_email='stefan.richter@cern.ch',
    description='',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.cern.ch/strichte/heppy',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Physics',
    ],
    install_requires=[
        'matplotlib',
        'numpy',
        'uproot',
        'uproot-methods',
        'awkward',
        'pandas',
        'pyyaml',
        'sphinx_rtd_theme',
    ],
    python_requires='>=3.6',
)



import numpy as np
import yoda

# Read the names of all the yoda histograms in the file
def get_histogram_names(yodafile):
    aos = yoda.read(yodafile, asdict=False)
    return [ao.name for ao in aos if ao.type == 'Histo1D']

# Read the paths of all the yoda histograms in the file
def get_histogram_paths(yodafile):
    aos = yoda.read(yodafile, asdict=False)
    return [ao.path for ao in aos if ao.type == 'Histo1D']

# Needs a reference yoda file containing the nominal event counter to whose sum of weights we want to normalise
# This is needed to undo the Athena Sherpa interface's patch/hack where it rescales the variation weights such
# that correct normalisation is obtained by taking (variation_weight_i) / (sum_of_NOMINAL_weights).
# Usually: refyodafile = path to file containing the nominal histograms!
def get_bin_heights(histogram_path, yodafile, refyodafile, debug=False):
    aos = yoda.read(yodafile, asdict=True)
    refaos = yoda.read(refyodafile, asdict=True)
    sow = aos['/_EVTCOUNT'].sumW()
    refsow = refaos['/_EVTCOUNT'].sumW() # in ATLAS usage, this must be the *nominal* sum of weights for the sample
    weight_correction = sow / refsow
    if debug:
        print('Correcting variation weights by {0} (variation SoW: {1}, reference SoW: {2})'.format(weight_correction, sow, refsow))
    return [b.height * weight_correction for b in aos[histogram_path].bins]

def get_curve(histogram_path, yodafile, refyodafile):
    heights = np.array(get_bin_heights(histogram_path, yodafile, refyodafile))
    return np.repeat(heights, 2)

def get_curves(histogram_path, yodafiles, refyodafile):
    return [get_curve(histogram_path, yodafile, refyodafile) for yodafile in yodafiles]

def get_relative_error(histogram_path, yodafile):
    aos = yoda.read(yodafile, asdict=True)
    return np.repeat(np.array([b.relErr for b in aos[histogram_path].bins]), 2)

def get_bin_edges(histogram_path, yodafile):
    binedges = []
    aos = yoda.read(yodafile, asdict=True)
    for b in aos[histogram_path].bins:
        binedges += list(b.xEdges)
    return binedges
